use std::{
    error::Error,
    fs::{read, File},
    path::{Path, PathBuf},
    sync::mpsc::{self, Receiver, Sender},
    thread::{self, JoinHandle},
};

use brotli::enc::{backward_references::BrotliEncoderMode, BrotliEncoderParams};
use clap::Parser;
use rusqlite::Connection;
use walkdir::WalkDir;

#[derive(Parser, Debug)]
#[command(author, version, about, long_about = None)]
struct Args {
    /// A donde guardar el archivo
    #[arg(short, long)]
    output: String,

    /// Carpeta a guardar
    #[arg(short, long)]
    input: String,
}

fn main() -> Result<(), Box<dyn Error>> {
    let args = Args::parse();
    {
        let conn = make_conn(&args.output)?;
        make_db_schema(&conn)?;
    }
    let input = Path::new(&args.input).to_owned();
    let paths = iterator(&input).collect::<Vec<PathBuf>>();

    let n_threads = 4;
    let chunks = paths.chunks(paths.len() / n_threads + paths.len() % n_threads);
    println!("threads: {}", chunks.len());

    let (rx, threads) = {
        let (tx, rx): (Sender<Entry>, Receiver<Entry>) = mpsc::channel();
        let producers = chunks
            .map(|chunk| {
                let chunkk = chunk.to_vec();
                let ttx = tx.clone();
                let inputt = input.clone();
                thread::spawn(move || producer(&inputt, chunkk, ttx))
            })
            .collect::<Vec<JoinHandle<()>>>();

        (rx, producers)
    };
    collector(&args.output, rx);
    for thread in threads {
        thread.join().unwrap();
    }
    Ok(())
}

fn producer(input: &Path, chunkk: Vec<PathBuf>, ttx: Sender<(String, Vec<u8>, bool)>) {
    let params = file_compress_params();
    for path in chunkk {
        let tx = &ttx;
        let path_strip = strip_input(&path, input);
        let path_str = path_strip.to_string_lossy();
        if compressable(&path) {
            let mut content_compressed = Vec::new();
            let mut file = File::open(&path).unwrap();
            brotli::BrotliCompress(&mut file, &mut content_compressed, &params).unwrap();
            tx.send((path_str.to_string(), content_compressed, true))
                .unwrap();
        } else {
            let content = read(path.clone()).unwrap();
            tx.send((path_str.to_string(), content, false)).unwrap();
        }
    }
}

fn collector(output: &str, rx: Receiver<(String, Vec<u8>, bool)>) {
    let mut conn = make_conn(output).unwrap();
    let trans = conn.transaction().unwrap();
    {
        let mut insert_stmt = trans
            .prepare_cached("insert into files(path, content, compressed) values(?, ?, ?)")
            .unwrap();
        for entry in rx {
            insert_stmt.execute(entry).unwrap();
        }
    }
    trans.commit().unwrap();
}

fn make_conn(output: &str) -> Result<Connection, Box<dyn Error>> {
    let conn = Connection::open(output)?;
    conn.pragma_update(None, "journal_mode", "OFF")?;
    conn.pragma_update(None, "synchronous", 0)?;
    conn.pragma_update(None, "cache_size", 1000000)?;
    conn.pragma_update(None, "locking_mode", "EXCLUSIVE")?;
    Ok(conn)
}

type Entry = (String, Vec<u8>, bool);

fn make_db_schema(conn: &Connection) -> Result<(), Box<dyn Error>> {
    conn.execute(
        "create table files(path text, content blob, compressed bool)",
        [],
    )?;
    conn.execute("create unique index path on files(path)", [])?;
    Ok(())
}

fn file_compress_params() -> BrotliEncoderParams {
    let mut params = BrotliEncoderParams::default();
    params.quality = 0;
    params.mode = BrotliEncoderMode::BROTLI_MODE_TEXT;
    params
}

fn iterator(input: &Path) -> impl Iterator<Item = PathBuf> {
    WalkDir::new(input)
        .into_iter()
        .filter_map(|e| e.ok())
        .filter(|e| e.file_type().is_file())
        .filter(|e| !already_compressed(e.path()))
        .map(|e| e.path().to_owned())
}

fn strip_input(path: &Path, input: &Path) -> PathBuf {
    path.strip_prefix(input).unwrap().to_owned()
}

fn already_compressed(path: &Path) -> bool {
    let p = path.to_string_lossy();
    p.ends_with(".br") || p.ends_with(".gz")
}

fn compressable(path: &Path) -> bool {
    let p = path.to_string_lossy();
    p.ends_with(".html")
        || p.ends_with(".css")
        || p.ends_with(".js")
        || p.ends_with(".json")
        || p.ends_with(".svg")
}
